#include "config.h"

#include <stdio.h>
#include "tcpserver.h"

// Near-useless constructor, allthe work is done in Setup(...)
TcpServer::TcpServer()
{
    sv_valid = 0;

    // Set the serv_fd to 0 as well so we don't close() an unopen sock
    serv_fd = 0;

    max_fd = 0;
}

// The deconstructor actually does some good work.
TcpServer::~TcpServer()
{
    // Invalidate us immediately
    sv_valid = 0;

    // and close down all our sockets
    if (serv_fd)
        close(serv_fd);
}

void TcpServer::Shutdown() {
    // Invalidate us immediately
    sv_valid = 0;

    // and close down all our sockets
    if (serv_fd)
        close(serv_fd);

}

// Bind to a port and optional hostname/interface
int TcpServer::Setup(unsigned int in_max_clients, short int in_port, const char *in_allowed)
{
    max_clients = in_max_clients;
    allowed = in_allowed;

    // If we don't have a host to bind to, try to find one
    /*
     if (in_host == NULL) {
     */
    // Die violently -- If we can't bind a socket, we're useless
    if (gethostname(hostname, MAXHOSTNAMELEN) < 0) {
        snprintf(errstr, 1024, "TcpServer gethostname() failed: %s", strerror(errno));
        return (-1);
    }
    /*
    } else {
        // Use the one provided
        strncpy(hostname, in_host, MAXHOSTNAMELEN);
        }
        */

    // Copy the port to our local data
    port = in_port;

    // Resolve the hostname we were given/found to see if it's actually
    // valid
    if ((serv_host = gethostbyname(hostname)) == NULL) {
        snprintf(errstr, 1024, "TcpServer gethostbyname() failed: %s", strerror(errno));
        return (-2);
    }

    // Set up our socket
    bzero(&serv_sock, sizeof(serv_sock));
    serv_sock.sin_family = AF_INET;
    serv_sock.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_sock.sin_port = htons(port);

    // Debug("Server::Setup calling socket()");
    if ((serv_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        snprintf(errstr, 1024, "TcpServer socket() failed: %s", strerror(errno));
        return (-3);
    }

    // Debug("Server::Setup setting socket option SO_REUSEADDR");
    int i = 2;
    if (setsockopt(serv_fd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) == -1) {
        snprintf(errstr, 1024, "TcpServer setsockopt() failed: %s", strerror(errno));
        return (-4);
    }

    // Bind the named socket
    // Debug("Server::Setup calling bind()");
    if (bind(serv_fd, (struct sockaddr *) &serv_sock, sizeof(serv_sock)) < 0) {
        snprintf(errstr, 1024, "TcpServer bind() failed: %s", strerror(errno));
        return (-5);
    }

    // Start up listening to the socket
    if (listen(serv_fd, 5) < 0) {
        snprintf(errstr, 1024, "TcpServer listen() failed: %s", strerror(errno));
        return (-6);
    }

    // Zero the FD's
    // Debug("Server::Setup zero-ing FD sets");
    FD_ZERO(&server_fds);
    FD_ZERO(&read_fds);
    FD_ZERO(&except_fds);
    FD_ZERO(&client_fds);

    // Set the server socket
    FD_SET(serv_fd, &server_fds);
    sv_valid = 1;

    if (serv_fd > max_fd)
        max_fd = serv_fd;

    return (1);
}

// Make one useable fd_set from the fd's flagged for system-wide monitoring
// and from the fd's flagged locally for clients connecting to us.  This lets
// us do 1 big unified select().
fd_set TcpServer::MergeSet(fd_set in_set, unsigned int in_max) {
    unsigned int max;
    fd_set ret;

    FD_ZERO(&ret);

    if (in_max < max_fd)
        max = max_fd;
    else
        max = in_max;

    for (unsigned int x = 0; x <= max; x++) {
        if (FD_ISSET(x, &in_set) || FD_ISSET(x, &server_fds))
            FD_SET(x, &ret);
    }

    return ret;
}

int TcpServer::Poll(fd_set in_rset, fd_set in_wset, fd_set in_eset)
{
    if (!sv_valid)
        return -1;

    // int selected;

    /*
    // Initialze our FD's
    read_fds = server_fds;

    struct timeval tm;
//    tm.tv_sec = TCP_SELECT_TIMEOUT / 1000;
//    tm.tv_usec = (TCP_SELECT_TIMEOUT % 1000) * 1000;

    tm.tv_sec = 0;
    tm.tv_usec = 0;

    // Enter the select loop
    //Debug("Server::Poll() - Calling select()");
    if ((selected = select(max_fd+1, &read_fds, NULL, &except_fds, &tm)) < 0) {
        // Set selected to 0 if we were interrupted
        if (errno != EINTR) {
            snprintf(errstr, 1024, "TcpServer select() failed: %s",
                     strerror(errno));
            return (-1);
        }
        }
        */

    int accept_fd = 0;
    if (FD_ISSET(serv_fd, &in_rset))
        if ((accept_fd = Accept()) < 0)
            return -1;

    for (unsigned int fd = serv_fd; fd < max_fd; fd++) {
        if (FD_ISSET(fd, &in_eset))
            Kill(fd);
    }

    return accept_fd;
}

// Accept an incoming connection
int TcpServer::Accept()
{
    unsigned int new_fd;
    struct sockaddr_in client_addr;
#ifdef HAVE_SOCKLEN_T
    socklen_t client_len;
#else
	int client_len;
#endif

    bzero(&client_addr, sizeof(struct sockaddr_in));

    client_len = sizeof(struct sockaddr_in);

    // Accept should never block, thanks to select
    if ((new_fd = accept(serv_fd, (struct sockaddr *) &client_addr,
                         &client_len)) < 0) {
        snprintf(errstr, 1024, "TcpServer accept() failed: %s\n", strerror(errno));
        return -1;
    }

    char inhost[16];

    snprintf(inhost, 16, "%s", inet_ntoa(client_addr.sin_addr));

    if (!strstr(allowed, inhost)) {
        snprintf(errstr, 1024, "TcpServer accept() connect from untrusted host %s",
                 inhost);
        close(new_fd);
        return -1;
    } else {
        snprintf(errstr, 1024, "%s", inhost);
    }

    if (new_fd > max_fd)
        max_fd = new_fd;

    // Set the file descriptor
    FD_SET(new_fd, &server_fds);
    FD_SET(new_fd, &client_fds);

    // Set it to nonblocking.  We only care about spewing data out to them.
    int save_mode = fcntl(new_fd, F_GETFL, 0);
    fcntl(new_fd, F_SETFL, save_mode | O_NONBLOCK);

    return new_fd;
}

// Kill a connection
void TcpServer::Kill(int in_fd)
{
    FD_CLR(in_fd, &server_fds);
    FD_CLR(in_fd, &client_fds);
    if (in_fd)
        close(in_fd);
}

void TcpServer::Send(int in_fd, const char *in_data) {
    if (write(in_fd, in_data, strlen(in_data)) < 0) {
        if (errno != EAGAIN)
            Kill(in_fd);
    }
}

void TcpServer::SendToAll(const char *in_data) {
    for (unsigned int x = serv_fd; x <= max_fd; x++) {
        if (!FD_ISSET(x, &client_fds))
            continue;

        Send(x, in_data);
    }
}

